--- Status: 27.03.2017

--- disable legacy SQL

--------------------------------------------------------
--- ANSWER POSTS
--------------------------------------------------------

--- Get answers to questions tagged with "java" or "android"
SELECT
  a.id as post_id,
  a.parent_id as parent_id,
  q.java_tag as java_tag,
  q.android_tag as android_tag,
  a.creation_date as creation_date,
  a.score as score
FROM
  `bigquery-public-data.stackoverflow.posts_answers` a
  INNER JOIN
  `so-snippets.so_java.questions` q
  ON q.post_id = a.parent_id;

=> so_java.answers

--- Number of answers
SELECT COUNT(DISTINCT id)
FROM `so-snippets.so_java.answers`;
=> 3,220,450


--- Goal: Get most recent version of java answers from post_history.
--- Advantages: * Markdown without HTML character escaping like in posts_answers.
---             * Differentiation between code blocks and inline code highlighting (code spans, see http://stackoverflow.com/editing-help#code).
--- Problem: BigQuery SO data set has broken whitespaces (in particular "    " for code blocks) for some posts (e.g. http://stackoverflow.com/a/20861931).
--- Solution: Export Answer IDs to CSV, import in MySQL with SO dump, get posts from posthistory, import in BigQuery.
SELECT post_id
FROM `so-snippets.so_java.answers`;

=> so_java.answers_export

--- import java_answers_text.csv from MySQL
=> so_java.answers_text


--- Compare number of extracted post_ids with table post_answers and check whether only one row per post_id exists.
SELECT COUNT(*)
FROM `so-snippets.so_java.answers`;
=> 3,220,450

SELECT COUNT(DISTINCT post_id)
FROM `so-snippets.so_java.answers`;
=> 3,220,450

SELECT COUNT(*)
FROM `so-snippets.so_java.answers_text`;
=> 3,220,448

SELECT COUNT(DISTINCT post_id)
FROM `so-snippets.so_java.answers_text`;
=> 3,220,448


--- analyze date difference between creation and edit date
SELECT *
FROM `so-snippets.so_java.answers_text`
WHERE DATE_DIFF(CAST(last_edit_date AS DATE), CAST(creation_date AS DATE), DAY) > 1;


--------------------------------------------------------
--- ANSWER SNIPPETS
--------------------------------------------------------

--- extract code blocks from answers
SELECT
  post_id,
  FARM_FINGERPRINT(TRIM(code_block)) as code_block_hash, --- hash each code block (-> identifier)
  ARRAY_LENGTH(SPLIT(TRIM(code_block), '\n')) as line_count, --- get number of lines in codeblock
  TRIM(code_block) as code_block --- trim is important to remove empty lines at the beginning or end of codeblocks
FROM (
  SELECT
    post_id,
    code_blocks
  FROM (
    SELECT
      post_id,
      ARRAY(
        SELECT
          code_block_line
        FROM
          UNNEST(
            SPLIT(
              REGEXP_REPLACE(
                ARRAY_TO_STRING(lines, '\n'), --- merge lines of code blocks and replacement for text blocks
                r'(a850f1c3a905a260(?:\na850f1c3a905a260)+)', "a850f1c3a905a260" --- reduce consecutive lines with text block marker to one marker
              ),  
            'a850f1c3a905a260') --- split resulting string into one array element per code block
          ) as code_block_line
        WHERE
          LENGTH(code_block_line) > 0 AND NOT REGEXP_CONTAINS(code_block_line, r'^\s*$') --- ignore empty and whitespace lines again
      ) as code_blocks
    FROM (
      SELECT
        post_id,
        ARRAY( --- replace lines with text by a marker pattern
          SELECT
            CASE
              WHEN STARTS_WITH(line, '    ') OR STARTS_WITH(line, '\t') --- tab is important here (see, e.g., 19518309)
              THEN line --- code block
              ELSE 'a850f1c3a905a260' --- text block
            END
          FROM
            UNNEST(lines) as line
          WHERE
            LENGTH(line) > 0 AND NOT REGEXP_CONTAINS(line, r'^\s*$')  --- ignore empty lines and lines only consisting of whitespaces to prevent accidential split of code blocks
        ) as lines
      FROM (
        SELECT
          post_id,
          SPLIT(
            REGEXP_REPLACE(
              REGEXP_REPLACE(text, r'&#xD;&#xA;', '\n'), --- replace HTML-encoded "\r\n"
              r'\r*\n+', '\n'), --- normalize line breaks
            '\n') as lines --- split answers into lines
        FROM
          `so-snippets.so_java.answers_text`
      )
    )
  )
  WHERE ARRAY_LENGTH(code_blocks) > 0 --- ignore posts without code blocks
) as posts
CROSS JOIN
UNNEST(posts.code_blocks) as code_block; --- flatten array to rows

=> so_java.answers_text_code_blocks
    

--- get number of code blocks and check for inconsistencies
SELECT
  COUNT(*)
FROM
  `so-snippets.so_java.answers_text_code_blocks`;
=> 2,644,206

SELECT
  COUNT(*)
FROM (
  SELECT
    post_id,
    code_block_hash
  FROM
    `so-snippets.so_java.answers_text_code_blocks`
  GROUP BY post_id, code_block_hash
);
=> 2,640,979
--- OBSERVATION: Seems like some snippets are present in different posts.


--- get codeblocks that are present in multiple posts
SELECT
  code_block_hash,
  COUNT(post_id>1) as post_count,
  ARRAY_AGG(post_id) as post_ids,
  code_block
FROM
  `so-snippets.so_java.answers_text_code_blocks`
GROUP BY
   code_block_hash, code_block
HAVING
  post_count > 1
ORDER BY
  post_count DESC;

 => so_java.answers_text_code_blocks_duplicates
  
--- TODO: Analyze duplicates within SO

  
--- normalize answer code blocks
--- TODO: also remove modifiers (static, abstract, public, private, protected, etc.) 
SELECT
  post_id,
  code_block_hash, --- trim is important to get rid of empty lines at the beginning and/or end
  ARRAY_LENGTH(SPLIT(TRIM(ARRAY_TO_STRING(lines, ' ')), ' ')) as line_count,
  TRIM(ARRAY_TO_STRING(lines, '')) as code_block --- merge all lines into one string whithout whitespaces
FROM (
  SELECT
    post_id,
    code_block_hash,
    SPLIT(
      REGEXP_REPLACE(
        REGEXP_REPLACE(
          ARRAY_TO_STRING(lines, ' '), --- reduce snippet to one line, separated by blanks
              r'(\/\*.*?\*\/)', r''), --- remove multiline comments 
              r'\s+', ' '), --- normalize whitespaces again
      " " --- split again for line count after normalization
    ) as lines
  FROM (
    SELECT
      post_id,
      code_block_hash,
      ARRAY(
        SELECT 
          line
        FROM
          UNNEST(lines) as line
        WHERE
          LENGTH(line) > 0 AND NOT REGEXP_CONTAINS(line, r'^\s*$')  --- ignore empty lines and lines only consisting of whitespaces
      ) as lines
    FROM (
      SELECT
        post_id,
        code_block_hash,
        ARRAY(
          SELECT
            REGEXP_REPLACE(
              REPLACE( --- remove special characters {, }, (, ), and ;
                REPLACE(
                  REPLACE(
                    REPLACE(
                      REPLACE( 
                        REGEXP_REPLACE(line, r'(^.*?)(\/\/.*)$', r'\1'), --- remove line comments
                      '{', ''),
                      '}', ''),
                      ';', ''),
                      '(', ''),
                      ')', ''),
              r'\s+', '' -- remove whitespaces
            )
          FROM 
            UNNEST(lines) as line
          WHERE
            NOT REGEXP_CONTAINS(line, r'^\s*import') --- remove import statements
            AND NOT REGEXP_CONTAINS(line, r'^\s*package') --- remove package declarations
            AND NOT REGEXP_CONTAINS(line, r'^\s*\.+\s*$') --- remove lines like "..."
        ) as lines
      FROM (
        SELECT
          post_id,
          code_block_hash,
          SPLIT(LOWER(code_block), '\n') as lines
        FROM
          (SELECT
            post_id,
            code_block_hash,
            code_block
          FROM
            `so-snippets.so_java.answers_text_code_blocks`
          WHERE
            REGEXP_CONTAINS(code_block, r';') --- at least one semicolon
            AND REGEXP_CONTAINS(code_block, r'\(') --- at least one regular brace
            AND REGEXP_CONTAINS(code_block, r'\{') --- at least one curly brace
            AND NOT REGEXP_CONTAINS(code_block, r'^\/?\s*<[^>]+>.*') --- remove XML files
            AND NOT REGEXP_CONTAINS(code_block, r'^\s*(?:\$|#)') --- remove PHP, JavaScript, Bashscript, etc.
          )
      )
    )
  )
)
WHERE LENGTH(TRIM(ARRAY_TO_STRING(lines, ''))) > 0;  --- ignore empty snippets (after normalization)

=> so_java.answers_text_code_blocks_normalized


---- get number of normalized code blocks and check for inconsistencies
SELECT
  COUNT(*)
FROM
  `so-snippets.so_java.answers_text_code_blocks_normalized`;
=> 1,063,993

SELECT
  COUNT(*)
FROM (
  SELECT
    post_id,
    code_block_hash
  FROM
    `so-snippets.so_java.answers_text_code_blocks_normalized`
  GROUP BY post_id, code_block_hash
);
=> 1,063,790
--- OBSERVATION: seems like some snippets are present in different posts


--- get codeblocks that are present in multiple posts
SELECT
  code_block_hash,
  COUNT(post_id) as post_count,
  ARRAY_AGG(post_id) as post_ids,
  code_block
FROM
  `so-snippets.so_java.answers_text_code_blocks_normalized`
GROUP BY
   code_block_hash, code_block
HAVING
  post_count > 1
ORDER BY
  post_count DESC;

 => so_java.answers_text_code_blocks_normalized_duplicates
    
--- TODO: Analyze duplicates within SO


--- create table without content for analasis in R
SELECT
  post_id,
  code_block_hash,
  line_count
FROM
  `so-snippets.so_java.answers_text_code_blocks_normalized`;
  
=> so_java.answers_text_code_blocks_normalized_size

--- Export and analyze in R
--- DECISION: ignore snippets with less than 6 normalized lines (7 is median value)


--- analyze score of normalized snippets
SELECT
  DISTINCT(an.post_id) as post_id, --- posts may have more than one snippet -> count only once
  a.score as score
FROM
  `so-snippets.so_java.answers` as a
INNER JOIN
  `so-snippets.so_java.answers_text_code_blocks_normalized` as an
ON
  a.post_id = an.post_id;
  
=> so_java.answers_text_code_blocks_normalized_score

--- Export and analyze in R
--- DECISION: ignore snippets with a score of less than 10
--- Other aspect: complexity of resulting combinations


--- create sample of non-trivial SO snippets
SELECT
  DISTINCT(code_block_hash) as code_block_hash, --- some snippets are present in multiple posts, can be joined again later
  line_count,
  code_block
FROM
  `so-snippets.so_java.answers` as a
INNER JOIN
  `so-snippets.so_java.answers_text_code_blocks_normalized` as an
ON
  a.post_id = an.post_id
WHERE
  an.line_count >= 6 --- exclude small snippets (problem: false positives)
  AND a.score >= 10; --- exclude downvoted, ignored, and unpopular snippets
  
=> so_java.answers_sample

--- count number of snippets in sample
SELECT
  COUNT(*)
FROM
  `so-snippets.so_java.answers_sample`;
=> 29,370

--- count number of posts where snippets are taken from
SELECT
  COUNT(DISTINCT(a.post_id)) as post_id_count
FROM
  `so-snippets.so_java.answers` as a
INNER JOIN
  `so-snippets.so_java.answers_text_code_blocks_normalized` as an
ON
  a.post_id = an.post_id
WHERE
  an.line_count >= 6 --- exclude small snippets (problem: false positives)
  AND a.score >= 10; --- exclude downvoted, ignored, and unpopular snippets

=> 23,829

--- divide into four samples to reduce complexity
SELECT *
FROM `so-snippets.so_java.answers_sample`
WHERE MOD(ABS(code_block_hash), 4) = 0;

=> so_java.answers_sample_1

SELECT COUNT(*)
FROM `so-snippets.so_java.answers_sample_1`;

=> 7,282


SELECT *
FROM `so-snippets.so_java.answers_sample`
WHERE MOD(ABS(code_block_hash), 4) = 1;

=> so_java.answers_sample_2

SELECT COUNT(*)
FROM `so-snippets.so_java.answers_sample_2`;

=> 7,423


SELECT *
FROM `so-snippets.so_java.answers_sample`
WHERE MOD(ABS(code_block_hash), 4) = 2;

=> so_java.answers_sample_3

SELECT COUNT(*)
FROM `so-snippets.so_java.answers_sample_3`;

=> 7,346
 

SELECT *
FROM `so-snippets.so_java.answers_sample`
WHERE MOD(ABS(code_block_hash), 4) = 3;

=> so_java.answers_sample_4

SELECT COUNT(*)
FROM `so-snippets.so_java.answers_sample_4`;

=> 7,319


==> 29,370 = 7,282 + 7,423 + 7,346 + 7,319

