--- Status: 23.03.2017

--- disable legacy SQL

--------------------------------------------------------
--- FILES
--------------------------------------------------------

--- get java files with content
SELECT
  f.id as id,
  f.repo_name as repo_name,
  f.path as path,
  c.content as content
FROM
  `so-snippets.gh.files` AS f
  JOIN `bigquery-public-data.github_repos.contents` AS c
  ON f.id = c.id
WHERE
  LOWER(path) LIKE '%.java';

=> gh_java.files_content
 

--- analyze file count per java project to exclude "toy" projects
SELECT repo_name, COUNT(DISTINCT id) as file_count
FROM `so-snippets.gh_java.files_content`
GROUP BY repo_name;

=> gh_java.repos_file_count

--- DECISION: ignore repos with less than 5 files (4 is 25% quantile)


--- analyze watcher count for java projects to exclude "toy" projects
SELECT repos.repo_name as repo_name, watcher_count
FROM (
  SELECT DISTINCT(repo_name)
  FROM `so-snippets.gh_java.files_content`
) as java_repos
INNER JOIN `so-snippets.gh.repos` as repos
ON java_repos.repo_name = repos.repo_name;

=> gh_java.repos_watcher_count

--- DECISION: ignore repos with less than 1 watcher (0 is median value)


--- filter repos according to above criteria to exclude "toy" projects
SELECT w.repo_name as repo_name
FROM `so-snippets.gh_java.repos_watcher_count` w
INNER JOIN `so-snippets.gh_java.repos_file_count` f
ON w.repo_name = f.repo_name
WHERE
  file_count >= 5
  AND watcher_count >= 1; 

 => gh_java.repos_filtered

SELECT COUNT(*)
FROM `so-snippets.gh_java.repos_filtered`;
=> 72,117

--- get license of filtered repos
SELECT
  repos.repo_name as repo_name,
  license
FROM `so-snippets.gh_java.repos_filtered` repos
INNER JOIN `bigquery-public-data.github_repos.licenses` licenses
ON repos.repo_name = licenses.repo_name;

=> gh_java.repos_filtered_license
 

 --- filter files (only from above repos, only unique files)
SELECT id, content
FROM `so-snippets.gh_java.files_content` f
INNER JOIN `so-snippets.gh_java.repos_filtered` r
ON f.repo_name = r.repo_name
GROUP BY id, content --- only consider unique files in the following

=> gh_java.files_content_filtered

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_content_filtered`;
=> 6,910,792


--- normalize file contents
SELECT
  id,
  ARRAY_LENGTH(SPLIT(TRIM(ARRAY_TO_STRING(lines, ' ')), ' ')) as line_count,
  TRIM(ARRAY_TO_STRING(lines, '')) as content --- merge all lines into one string whithout whitespaces
FROM (
  SELECT
    id,
    SPLIT(
      REGEXP_REPLACE(
        REGEXP_REPLACE(
          ARRAY_TO_STRING(lines, ' '), --- reduce snippet to one line, separated by blanks
              r'(\/\*.*?\*\/)', r''), --- remove multiline comments 
              r'\s+', ' '), --- normalize whitespaces again
      " " --- split again for line count after normalization
    ) as lines
  FROM (
    SELECT
      id,
      ARRAY(
        SELECT 
          line
        FROM
          UNNEST(lines) as line
        WHERE
          LENGTH(line) > 0 AND NOT REGEXP_CONTAINS(line, r'^\s*$')  --- ignore empty lines and lines only consisting of whitespaces
      ) as lines
    FROM (
      SELECT
        id,
        ARRAY(
          SELECT
            REGEXP_REPLACE(
              REPLACE( --- remove special characters {, }, (, ), and ;
                REPLACE(
                  REPLACE(
                    REPLACE(
                      REPLACE( 
                        REGEXP_REPLACE(line, r'(^.*?)(\/\/.*)$', r'\1'), --- remove line comments
                      '{', ''),
                      '}', ''),
                      ';', ''),
                      '(', ''),
                      ')', ''),
              r'\s+', '' -- remove whitespaces
            )
          FROM 
            UNNEST(lines) as line
          WHERE
            NOT REGEXP_CONTAINS(line, r'^\s*import') --- remove import statements
            AND NOT REGEXP_CONTAINS(line, r'^\s*package') --- remove package declarations
            AND NOT REGEXP_CONTAINS(line, r'^\s*\.+\s*$') --- remove lines like "..."
        ) as lines
      FROM (
        SELECT
          id,
          SPLIT(LOWER(content), '\n') as lines
        FROM
          `so-snippets.gh_java.files_content_filtered`
        WHERE
          REGEXP_CONTAINS(content, r'(class\s+|interface\s+|enum\s+)') --- Java files must contain class, interface, or enum definition
      )
    )
  )
)
WHERE
  LENGTH(TRIM(ARRAY_TO_STRING(lines, ''))) > 0;  --- ignore empty files (after normalization);

=> gh_java.files_content_filtered_normalized


--- get number of normalized java files
SELECT COUNT(*)
FROM `so-snippets.gh_java.files_content_filtered_normalized`;
=> 6,851,022


--- create table without content for analasis in R
SELECT id, line_count
FROM `so-snippets.gh_java.files_content_filtered_normalized`;

=> gh_java.files_content_filtered_normalized_size

--------------------------------------------------------
---  DECISION: ignore files with less than 68 normalized lines (68 is 75% quantile and 69.85038 is mean value)
--------------------------------------------------------


--- get sample of Java files for snippet search
SELECT *
FROM `so-snippets.gh_java.files_content_filtered_normalized`
WHERE line_count >= 68;

=> gh_java.files_sample

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample`;

=> 1,720,587


--- split files into four samples to reduce complexity of snippet search

SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 0;

=> gh_java.files_sample_1

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_1`;

=> 215,233


SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 1;

=> gh_java.files_sample_2

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_2`;

=> 214,464


SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 2;

=> gh_java.files_sample_3

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_3`;

=> 215,445


SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 3;

=> gh_java.files_sample_4

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_4`;

=> 215,160


SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 4;

=> gh_java.files_sample_5

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_5`;

=> 214,802


SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 5;

=> gh_java.files_sample_6

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_6`;

=> 215,237


SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 6;

=> gh_java.files_sample_7

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_7`;

=> 215,242


SELECT *
FROM `so-snippets.gh_java.files_sample`
WHERE MOD(ABS(FARM_FINGERPRINT(id)), 8) = 7;

=> gh_java.files_sample_8

SELECT COUNT(*)
FROM `so-snippets.gh_java.files_sample_8`;

=> 215,004


=> 1,720,587 = 215,233 + 214,464 + 215,445 + 215,160 + 214,802 + 215,237 + 215,242 + 215,004

