--- Status: 31.03.2017

--- disable legacy SQL

--------------------------------------------------------
--- CHECK ATTRIBUTION
--------------------------------------------------------

--- analyzes matches from answers with references to other SO posts or external sources

--- join with links to external resources
SELECT
  code_block_hash,
  ARRAY_TO_STRING(ARRAY_AGG(DISTINCT so_link), " ") AS so_links,
  ARRAY_TO_STRING(ARRAY_AGG(DISTINCT link), " ") AS links_from_post,
  code_block
FROM `so-snippets.clones_java.files_matches_complete_data` as matches
INNER JOIN `so-snippets.so_java.answers_text_links_not_so` as links
ON matches.post_id = links.post_id
GROUP BY code_block_hash, code_block;

=> attribution_java.files_matches_external_links

SELECT COUNT(*)
FROM `so-snippets.attribution_java.files_matches_external_links`;

=> 534

--- manually coded all code blocks with referenceds to external sources
=> attribution_java.files_matches_external_links_coded
--- DECISION: exclude code blocks from "android documentation, android source code, java documentation, and jdk source code" 
--- DECISION: separate analysis for github gist or repo of same/other author, source code hosted somewhere else or posted on another website

SELECT COUNT(*)
FROM `so-snippets.attribution_java.files_matches_external_links_coded`;

=> 534


--- separate code blocks having external references
SELECT *
FROM `so-snippets.attribution_java.files_matches_external_links_coded`
WHERE copied_from_external_source  = TRUE;

=> attribution_java.files_matches_external_links_copied

SELECT COUNT(*)
FROM `so-snippets.attribution_java.files_matches_external_links_copied`;

=> 350


SELECT *
FROM `so-snippets.attribution_java.files_matches_external_links_coded`
WHERE comment LIKE "post%" OR comment LIKE "github%" OR comment LIKE "source code%";

=> attribution_java.files_matches_external_links_other_source

SELECT COUNT(*)
FROM `so-snippets.attribution_java.files_matches_external_links_other_source`;

=> 159


--- remove matches with copies of snippets from android or JDK source code or documentation
SELECT
  file_id,
  repo_name,
  path,
  gh_link,
  match_pos,
  post_id,
  parent_id,
  post_history_id,
  so_link,
  creation_date,
  last_edit_date, 
  matches.code_block_hash as code_block_hash, 
  code_block
FROM `so-snippets.clones_java.files_matches_complete_data` matches
LEFT JOIN `so-snippets.attribution_java.files_matches_external_links_copied` filter
ON matches.code_block_hash = filter.code_block_hash
WHERE filter.code_block_hash IS NULL; --- workaround for missing minus operator

=> attribution_java.files_matches_filtered

SELECT COUNT(*)
FROM (
  SELECT file_id, code_block_hash
  FROM `so-snippets.attribution_java.files_matches_filtered`
  GROUP BY file_id, code_block_hash
);

=> 7,068

--- OBSERVATION: Still many matched files from mirrors/copies of JDK.
--- Search for code blocks responsible for these matches and exclude them.

SELECT code_block_hash, MIN(so_link) as so_link, code_block
FROM `so-snippets.attribution_java.files_matches_filtered`
WHERE LOWER(repo_name) LIKE "%jdk%"
  OR LOWER(path) LIKE "%jdk%"
  OR LOWER(path) LIKE "%java/lang/%"
  OR LOWER(path) LIKE "%android/platform/%"
  OR LOWER(path) LIKE "%com/sun/%"
GROUP BY code_block_hash, code_block;

=> java_files_matches_filtered_jdk_matches.csv

--- manually code matches, import java_files_matches_filtered_jdk_matches_coded.csv 

--- remove matches from JDK (manual coding and repo/path)
SELECT
  file_id,
  repo_name,
  path,
  gh_link,
  match_pos,
  post_id,
  parent_id,
  post_history_id,
  so_link,
  creation_date,
  last_edit_date, 
  matches.code_block_hash as code_block_hash, 
  code_block
FROM `so-snippets.attribution_java.files_matches_filtered` matches
LEFT JOIN (
  SELECT *
  FROM `so-snippets.attribution_java.files_matches_filtered_jdk_matches_coded`
  WHERE copied_from_jdk = FALSE
) filter
ON matches.code_block_hash = filter.code_block_hash
WHERE filter.code_block_hash IS NULL --- workaround for missing MINUS operator
  AND (NOT LOWER(repo_name) LIKE "%jdk%")
  AND (NOT LOWER(path) LIKE "%jdk%")
  AND (NOT LOWER(path) LIKE "%java/lang/%")
  AND (NOT LOWER(path) LIKE "%android/platform/%")
  AND (NOT LOWER(path) LIKE "%com/sun/%");

=> attribution_java.files_matches_filtered_tmp
=> attribution_java.files_matches_filtered

SELECT COUNT(*)
FROM (
  SELECT file_id, code_block_hash
  FROM `so-snippets.attribution_java.files_matches_filtered`
  GROUP BY file_id, code_block_hash
);

=> 6,779


--- OBSERVATION: same matches have very similar code blocks, but different code block hashes
--- DECISION: merge matches using normalized code blocks
SELECT
  file_id,
  repo_name,
  path,
  gh_link,
  match_pos,
  matches.post_id as post_id,
  parent_id,
  post_history_id,
  so_link,
  creation_date,
  last_edit_date, 
  matches.code_block_hash as code_block_hash, 
  FARM_FINGERPRINT(TRIM(norm.code_block)) as code_block_hash_norm,
  matches.code_block as code_block,
  norm.code_block as code_block_normalized
FROM `so-snippets.attribution_java.files_matches_filtered` matches
INNER JOIN `so-snippets.so_java.answers_text_code_blocks_normalized` norm
ON matches.post_id = norm.post_id AND matches.code_block_hash = norm.code_block_hash; 

=> attribution_java.files_matches_filtered_norm

--- number of remaining snippets
SELECT COUNT(DISTINCT code_block_hash_norm)
FROM `so-snippets.attribution_java.files_matches_filtered_norm`;
=> 909

SELECT COUNT(DISTINCT code_block_hash)
FROM `so-snippets.attribution_java.files_matches_filtered_norm`;
=> 852


--- search for other code blocks with (too) many matches (considering normalized code block hash)
SELECT
  code_block_hash_norm,
  COUNT(DISTINCT file_id) as file_count,
  ARRAY_LENGTH(ARRAY_AGG(DISTINCT so_link)) as so_link_count,
  ARRAY_TO_STRING(ARRAY_AGG(DISTINCT so_link), " ") as so_links,
  MIN(code_block) as code_block
FROM `so-snippets.attribution_java.files_matches_filtered_norm`
GROUP BY code_block_hash_norm
ORDER BY file_count DESC, so_link_count DESC;

=> java_files_matches_filtered_norm_file_count.csv


--- manually coded matches
--- only chose normalized code blocks coded as "ok"
--- import as "attribution_java.files_matches_filtered_norm_file_count_coded"
SELECT
  file_id,
  repo_name,
  path,
  gh_link,
  match_pos,
  post_id,
  parent_id,
  post_history_id,
  so_link,
  CAST(creation_date AS TIMESTAMP) as creation_date,
  last_edit_date, 
  code_block_hash, 
  matches.code_block_hash_norm as code_block_hash_norm,
  code_block,
  code_block_normalized
FROM `so-snippets.attribution_java.files_matches_filtered_norm` matches
INNER JOIN `so-snippets.attribution_java.files_matches_filtered_norm_file_count_coded` filter
ON matches.code_block_hash_norm = filter.code_block_hash_norm
WHERE TRIM(coding) = "ok";

=> attribution_java.files_matches_filtered_norm_filtered
=> 3,652

SELECT COUNT(*)
FROM (
  SELECT file_id, code_block_hash
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered`
  GROUP BY file_id, code_block_hash
);

=> 1,379


--- check which of the remaining matches are attributed
WITH
  matches as (
    SELECT
      file_id,
      post_id,
      parent_id,
      code_block_hash_norm
    FROM
      `so-snippets.attribution_java.files_matches_filtered_norm_filtered` m
    GROUP BY
      file_id, post_id, parent_id, code_block_hash_norm
  ),
  files_links_so_a as (
    SELECT
      file_id,
      REGEXP_EXTRACT(so_link, r'http:\/\/stackoverflow\.com\/a\/([\d]+)') as linked_a_id
    FROM
      `so-snippets.gh_java.files_content_links_so_aq`
    WHERE
      REGEXP_CONTAINS(so_link, r'http:\/\/stackoverflow\.com\/a\/([\d]+)')
    GROUP BY
      file_id, linked_a_id
  ),
  files_links_so_q as (
    SELECT
      file_id,
      REGEXP_EXTRACT(so_link, r'http:\/\/stackoverflow\.com\/q\/([\d]+)') as linked_q_id
    FROM
      `so-snippets.gh_java.files_content_links_so_aq`
    WHERE
      REGEXP_CONTAINS(so_link, r'http:\/\/stackoverflow\.com\/q\/([\d]+)')
    GROUP BY
      file_id, linked_q_id
  )
SELECT
  matches2.file_id as file_id,
  post_id,
  parent_id,
  code_block_hash_norm,
  CAST(linked_a_id as INT64) as linked_a_id,
  CAST(linked_q_id as INT64) as linked_q_id
FROM (
  SELECT
    matches.file_id as file_id,
    post_id,
    parent_id,
    code_block_hash_norm,
    linked_a_id
  FROM
    matches
  LEFT JOIN
    files_links_so_a
  ON
    matches.file_id = files_links_so_a.file_id
) as matches2
LEFT JOIN
  files_links_so_q
ON
  matches2.file_id = files_links_so_q.file_id;
  
=> attribution_java.files_matches_filtered_norm_filtered_links_so_aq
=> 3,057


--- for later validation
SELECT COUNT(*)
FROM (
  SELECT file_id, code_block_hash_norm
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq`
  GROUP BY file_id, code_block_hash_norm
);

=> 1,246


--- check attribution of filtered matches
SELECT
  matches.file_id as file_id,
  repo_name,
  path,
  gh_link,
  match_pos,
  matches.post_id as post_id,
  matches.parent_id as parent_id,
  post_history_id,
  so_link,
  creation_date,
  last_edit_date, 
  code_block_hash, 
  matches.code_block_hash_norm as code_block_hash_norm,
  code_block,
  code_block_normalized
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered` matches
INNER JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq` links
ON matches.file_id = links.file_id
  AND matches.code_block_hash_norm = links.code_block_hash_norm
WHERE links.post_id = linked_a_id
  OR links.parent_id = linked_q_id;

=> attribution_java.files_matches_filtered_norm_filtered_links_so_aq_attributed
=> 284


--- get unattributed filterd matches
SELECT
  filtered.file_id as file_id,
  repo_name,
  path,
  gh_link,
  match_pos,
  filtered.post_id as post_id,
  filtered.parent_id as parent_id,
  post_history_id,
  so_link,
  creation_date,
  last_edit_date, 
  code_block_hash, 
  filtered.code_block_hash_norm as code_block_hash_norm,
  code_block,
  code_block_normalized
FROM (
  SELECT
    matches.file_id as file_id,
    matches.post_id as post_id,
    matches.parent_id as parent_id,
    matches.code_block_hash_norm as code_block_hash_norm,
    matches.linked_a_id as linked_a_id,
    matches.linked_q_id as linked_q_id
  FROM
    `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq` matches
  LEFT JOIN
    `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq_attributed` attributed
  ON
    matches.file_id = attributed.file_id
    AND matches.code_block_hash_norm = attributed.code_block_hash_norm
  WHERE
    attributed.file_id IS NULL --- workaround for missing MINUS operator
) as unattributed
INNER JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered` filtered
ON unattributed.file_id = filtered.file_id
  AND unattributed.post_id = filtered.post_id
  AND unattributed.parent_id = filtered.parent_id
  AND unattributed.code_block_hash_norm = filtered.code_block_hash_norm;

=> attribution_java.files_matches_filtered_norm_filtered_links_so_aq_not_attributed
=> 3,383

--- validation
SELECT COUNT(*)
FROM (
  SELECT file_id, code_block_hash_norm
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq_attributed`
  GROUP BY file_id, code_block_hash_norm
);
=> 100

SELECT COUNT(*)
FROM (
  SELECT file_id, code_block_hash_norm
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq_not_attributed`
  GROUP BY file_id, code_block_hash_norm
);
=> 1,146

=> 100 + 1,146 = 1,246


--- attribution ratio
SELECT COUNT(*)
FROM (
  SELECT file_id, repo_name, path, code_block_hash_norm
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq_attributed`
  GROUP BY file_id, repo_name, path, code_block_hash_norm
);
=> 189

SELECT COUNT(*)
FROM (
  SELECT file_id, repo_name, path, code_block_hash_norm
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq_not_attributed`
  GROUP BY file_id, repo_name, path, code_block_hash_norm
);
=> 1,477

=> 189/(189+1,477) = 11.3% attributed)

=> 1,477/(189+1,477) = 88.7% not attributed)


--- TODO: check attribution for copies of snippets also available on other websites / in other repos    


-------------------------------------------------------------------------------


--- OBSERVATION: some code blocks are present in multiple SO answers

--- TODO: check this with normalized code blocks

--- analyze how many matched duplicated SO snippets have been attributed inside SO
WITH
  posts AS (SELECT
      *
    FROM
      `so-snippets.clones_java.files_matches_complete_data`),
  links_a AS (SELECT
      matches.post_id as post_id,
      REGEXP_EXTRACT(links.so_link, r'http:\/\/stackoverflow\.com\/a\/([\d]+)') as linked_a_id
    FROM `so-snippets.clones_java.files_matches_complete_data` as matches
    INNER JOIN `so-snippets.so_java.answers_text_links_so_aq` as links
    ON matches.post_id = links.post_id
    WHERE REGEXP_CONTAINS(links.so_link, r'http:\/\/stackoverflow\.com\/a\/([\d]+)')
    GROUP BY post_id, linked_a_id),
  links_q AS (SELECT
      matches.post_id as post_id,
      REGEXP_EXTRACT(links.so_link, r'http:\/\/stackoverflow\.com\/q\/([\d]+)') AS linked_q_id
    FROM `so-snippets.clones_java.files_matches_complete_data` as matches
    INNER JOIN `so-snippets.so_java.answers_text_links_so_aq` as links
    ON matches.post_id = links.post_id
    WHERE REGEXP_CONTAINS(links.so_link, r'http:\/\/stackoverflow\.com\/q\/([\d]+)')
    GROUP BY post_id, linked_q_id)
SELECT
  CAST(posts_a.post_id AS INT64) as post_id,
  CAST(posts_a.parent_id AS INT64) as parent_id,
  CAST(linked_a_id AS INT64) as linked_a_id,
  CAST(linked_q_id AS INT64) as linked_q_id
FROM (
  SELECT
    posts.post_id as post_id,
    parent_id,
    linked_a_id
  FROM
    posts
  LEFT JOIN
    links_a
  ON posts.post_id = links_a.post_id
) as posts_a
LEFT JOIN
  links_q
ON posts_a.post_id = links_q.post_id
WHERE CAST(posts_a.post_id AS INT64) = CAST(linked_a_id AS INT64)
  OR CAST(posts_a.parent_id AS INT64) = CAST(linked_q_id AS INT64);

=> attribution_java.files_matches_complete_data_text_links_so_aq_attributed

SELECT COUNT(DISTINCT post_id)
FROM `so-snippets.attribution_java.files_matches_complete_data_text_links_so_aq_attributed`;

=> 5

--- RESULT: Only 5 posts attributed the duplication
