--- Status: 02.04.2017

--- disable legacy SQL

--------------------------------------------------------
--- CHECK COMMIT DATE VS. CREATION DATE ON SO
--------------------------------------------------------


--- get distinct matches in files
SELECT DISTINCT repo_name, path, code_block, code_block_normalized
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered`
GROUP BY repo_name, path, code_block, code_block_normalized;

=> java_files_matches_filtered_norm_filtered_matches.csv

--- Use api-retriever to retrieve commits for matches in files, import into
--- attribution_java.files_matches_filtered_norm_filtered_matches_gh_data

--- REMARK: this tables contains data about matches that have been marked as "too trivial" in a later iteration of the coding process

--- merge matches and commit data
SELECT
  file_id,
  matches.repo_name as repo_name,
  matches.path as path,
  gh_link,
  match_pos,
  post_id,
  parent_id,
  post_history_id,
  so_link,
  creation_date,
  last_edit_date, 
  code_block_hash, 
  code_block_hash_norm,
  commit_sha,
  commit_date,
  code_block,
  matches.code_block_normalized as code_block_normalized
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered` matches
LEFT JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_matches_gh_data` commits
ON matches.repo_name = commits.repo_name
  AND matches.path = commits.path
  AND matches.code_block_normalized = commits.code_block_normalized
GROUP BY file_id, repo_name, path, gh_link, match_pos,
  post_id, parent_id, post_history_id, so_link, creation_date, last_edit_date, 
  code_block_hash, code_block_hash_norm, commit_sha, commit_date, code_block, code_block_normalized;
  
=> attribution_java.files_matches_filtered_norm_filtered_commits
=> 3,652 

--- TODO: analyze timespan between post on SO and commit on GH


--- We don't know from which post the GH user copied the snippet. Choose oldest thread on SO in the following.
SELECT
  file_id,
  repo_name,
  path,
  code_block_hash_norm,
  MIN(creation_date) as creation_date,
  MAX(commit_date) as commit_date
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits`
WHERE commit_date IS NOT NULL
GROUP BY file_id, repo_name, path, code_block_hash_norm

=> attribution_java.files_matches_filtered_norm_filtered_commits_unique


SELECT file_id, repo_name, path, code_block_hash_norm, creation_date, commit_date,
  DATE_DIFF(DATE(commit_date), DATE(creation_date), DAY) as diff_days
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_unique`
WHERE commit_date IS NOT NULL;

=> 918

--- get matches where commit is older than SO post
SELECT *
FROM (
  SELECT file_id, repo_name, path, code_block_hash_norm, creation_date, commit_date,
    DATE_DIFF(DATE(commit_date), DATE(creation_date), DAY) as diff_days
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_unique`
  WHERE commit_date IS NOT NULL
)
WHERE diff_days < 0;

=> 10
=> Either not copied from SO or from another post we did not consider in our analysis


SELECT
  filtered.file_id as file_id,
  filtered.repo_name as repo_name,
  filtered.path as path,
  filtered.code_block_hash_norm as code_block_hash_norm,
  CAST(filtered.creation_date AS DATETIME) as creation_date,
  CAST(filtered.commit_date AS DATETIME) as commit_date,
  gh_link,
  so_link,
  code_block
FROM (
  SELECT file_id, repo_name, path, code_block_hash_norm, creation_date, commit_date,
    DATE_DIFF(DATE(commit_date), DATE(creation_date), DAY) as diff_days
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_unique`
  WHERE commit_date IS NOT NULL
) filtered
LEFT JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits` matches
ON filtered.file_id = matches.file_id
  AND filtered.repo_name = matches.repo_name
  AND filtered.path = matches.path
  AND filtered.code_block_hash_norm = matches.code_block_hash_norm
  AND filtered.creation_date = matches.creation_date
  AND filtered.commit_date = matches.commit_date
WHERE diff_days < 0
GROUP BY file_id, repo_name, path, code_block_hash_norm,
  creation_date, commit_date, gh_link, so_link, code_block;

=> 10
=> files_matches_filtered_norm_filtered_commits_older


--- get distinct normalized code blocks for these matches
SELECT
  ARRAY_TO_STRING(ARRAY_AGG(DISTINCT gh_link), " ") as gh_links,
  ARRAY_TO_STRING(ARRAY_AGG(DISTINCT so_link), " ") as so_links,
  filtered.code_block_hash_norm as code_block_hash_norm,
  code_block,
  MAX(diff_days) as diff_days
FROM (
  SELECT file_id, repo_name, path, code_block_hash_norm, creation_date, commit_date,
    DATE_DIFF(DATE(commit_date), DATE(creation_date), DAY) as diff_days
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_unique`
  WHERE commit_date IS NOT NULL
) filtered
LEFT JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits` matches
ON filtered.file_id = matches.file_id
  AND filtered.repo_name = matches.repo_name
  AND filtered.path = matches.path
  AND filtered.code_block_hash_norm = matches.code_block_hash_norm
  AND filtered.creation_date = matches.creation_date
  AND filtered.commit_date = matches.commit_date
WHERE diff_days < 0
GROUP BY code_block_hash_norm, code_block;

=> 8
=> java_files_matches_filtered_norm_filtered_commits_older_code_blocks.csv


--- get matches where commit is newer than SO post (or same age)
SELECT
  filtered.file_id as file_id,
  filtered.repo_name as repo_name,
  filtered.path as path,
  filtered.code_block_hash_norm as code_block_hash_norm,
  CAST(filtered.creation_date AS DATETIME) as creation_date,
  CAST(filtered.commit_date AS DATETIME) as commit_date,
  gh_link,
  so_link,
  code_block
FROM (
  SELECT file_id, repo_name, path, code_block_hash_norm, creation_date, commit_date,
    DATE_DIFF(DATE(commit_date), DATE(creation_date), DAY) as diff_days
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_unique`
  WHERE commit_date IS NOT NULL
) filtered
LEFT JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits` matches
ON filtered.file_id = matches.file_id
  AND filtered.repo_name = matches.repo_name
  AND filtered.path = matches.path
  AND filtered.code_block_hash_norm = matches.code_block_hash_norm
  AND filtered.creation_date = matches.creation_date
  AND filtered.commit_date = matches.commit_date
WHERE diff_days >= 0
GROUP BY file_id, repo_name, path, code_block_hash_norm,
  creation_date, commit_date, gh_link, so_link, code_block;

=> 908
=> files_matches_filtered_norm_filtered_commits_newer


--- get distinct tuples (not normalized)
SELECT COUNT(*)
FROM (
  SELECT f.file_id as file_id, code_block_hash
  FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered` f 
  LEFT JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_older` o
  ON o.file_id = f.file_id
    AND o.repo_name = f.repo_name
    AND o.path = f.path
    AND o.code_block_hash_norm = f.code_block_hash_norm
  WHERE o.file_id IS NULL --- workaround for missing MINUS operator
  GROUP BY file_id, code_block_hash
);

files_matches_filtered_norm_filtered_commits_newer

=> 1,369


--- get repos with filtered matches
SELECT COUNT(DISTINCT repo_name)
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_newer`;

=> 764

SELECT COUNT(DISTINCT repo_name)
FROM `so-snippets.gh_java.repos_filtered`;

--- TODO: Wrong, 64,281 is correct
=> 64,281

=> 764/64,281 -> 1.19% of analyzed repos

SELECT DISTINCT repo_name
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_newer`;

=> java_files_matches_filtered_norm_filtered_commits_newer_repos.csv


-- attribution of matched files
SELECT COUNT(*)
FROM (
  SELECT m.file_id as file_id, m.code_block_hash as code_block_hash
  FROM (
    SELECT f.file_id as file_id, code_block_hash
    FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered` f 
    LEFT JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_older` o
    ON o.file_id = f.file_id
      AND o.repo_name = f.repo_name
      AND o.path = f.path
      AND o.code_block_hash_norm = f.code_block_hash_norm
    WHERE o.file_id IS NULL --- workaround for missing MINUS operator
    GROUP BY file_id, code_block_hash
  ) m
  INNER JOIN `so-snippets.attribution_java.files_matches_filtered_norm_filtered_links_so_aq_attributed` a
  ON m.file_id = a.file_id AND m.code_block_hash = a.code_block_hash
  GROUP BY file_id, code_block_hash
);

=> 104

=> 104/1,369 => 7.60%


--- analyze license of repos with matches
SELECT
  license,
  COUNT(DISTINCT matches.repo_name) as repo_count
FROM `so-snippets.attribution_java.files_matches_filtered_norm_filtered_commits_newer` matches
INNER JOIN `so-snippets.gh_java.repos_filtered_license` repos
ON matches.repo_name = repos.repo_name
GROUP BY license
ORDER BY repo_count DESC;

=> java_files_matches_filtered_norm_filtered_commits_newer_repos_license.csv












